// config.c

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <arpa/inet.h>

#include "fritzmon.h"
#include "config.h"

// Remove comment and new line characters
int remove_comment(char *sLine, char **sEnd)
{
	int i = 0;
	char cQuote = '\0';
	
	while (sLine[i]) {
		switch (sLine[i]) {
		case '#' :
			if (cQuote) // ignore if inside quoted string
				break;
		case '\r' : // accept also MS text files
		case '\n' : // terminate string and trigger end of loop
			sLine[i--] = '\0';
			break;
		case '\'' : // quotes
		case '"' :
			if (!cQuote)
				cQuote = sLine[i];
			// It's save to access sLine[i-1] here because
			// i is greater than 0 when cQuote matches.
			else if (cQuote == sLine[i] && sLine[i-1] != '\\') {
				cQuote = '\0';
				if (NULL != sEnd)
					*sEnd = sLine + i;
			}
			break;
		}
		++i;
	}
	return cQuote == '\0' ? 0 : -1;
}

// Parse line from configuration file.
// Modifies sLine to seperate option and value.
// Removes optional quotes from value.
// Returns option, passes value pointer to sValue.
char *parse_config(char *sLine, const char **sValue)
{
	char cStop;
	char *sEnd = NULL;
	char *sParam = NULL;
	char *s1;

	if (remove_comment(sLine, &sEnd)) { // wrong quoting (missing closing quote char)
		log_msg(FM_LOG_ERR, "config: Missing closing quote char");
		return NULL;
	}

	// Get option. Leading and trailing white spaces are ignored.
	sParam = sLine;
	while (isspace(*sParam))
		++sParam;
	s1 = sParam;
	while (*s1 && *s1 != '=' && !isspace(*s1))
		++s1;
	cStop = *s1;
	*s1++ = '\0';
	if (*sParam != '\0' && cStop != '=') {
		while (isspace(*s1))
			++s1;
		if (*s1++ != '=') {
			log_msg(FM_LOG_ERR, "config: Missing '=' char after option '%s'", sParam);
			sParam = NULL;
		}
	}
	if (NULL != sParam && *sParam != '\0') {
    		// Get value.
//		log_msg(FM_LOG_DEBUG, "config: Processing option %s", sParam);
		while (isspace(*s1))
			++s1;
		if (*s1 == '\'' || *s1 == '"') {
			if (NULL != sEnd) { // value is quoted, sEnd points to closing quote char
				++s1; // skip leading quote char
				*sEnd = '\0'; // terminate at trailing quote char
			}
			else // wrong quoting
				sParam = NULL;
		}
		else if (NULL != sEnd) // wrong quoting
			sParam = NULL;
		*sValue = s1;
	}
	return sParam;
}

// Get integer value and store in config var.
int set_int_config(const char *sOption, const char *sValue, int *cfg_var, int nMin, int nMax)
{
	int nResult = -1;
	int nVar;

	nVar = atoi(sValue);
	if (*sValue != '-' && !isdigit(*sValue))
		log_msg(FM_LOG_ERR, "config: Invalid parameter for option %s (not a number)", sOption);
	else if (nVar < nMin || nVar > nMax)
		log_msg(FM_LOG_ERR, "config: Invalid parameter for option %s (out of range)", sOption);
	else {
#ifdef DEBUG
		log_msg(FM_LOG_DEBUG, "config: %s = %d", sOption, nVar);
#endif
		*cfg_var = nVar;
		nResult = 0;
	}
	return nResult;
}

// Get boolean value and store in config var.
int set_bool_config(const char *sOption, const char *sValue, int *cfg_var)
{
	int nResult = 0;

	if (!strcmp(sValue, "1") ||
		!strcasecmp(sValue, "yes") ||
		!strcasecmp(sValue, "on") ||
		!strcasecmp(sValue, "true"))
		*cfg_var = 1;
	else if (!strcmp(sValue, "0") ||
		!strcasecmp(sValue, "no") ||
		!strcasecmp(sValue, "off") ||
		!strcasecmp(sValue, "false"))
		*cfg_var = 0;
	else {
		log_msg(FM_LOG_ERR, "config: Invalid parameter for option %s (not a boolean value)", sOption);
		nResult = -1;
	}
#ifdef DEBUG
	if (nResult == 0)
		log_msg(FM_LOG_DEBUG, "config: %s = %d", sOption, *cfg_var);
#endif
	return nResult;
}

// Get string value and store in config var.
int set_str_config(const char *sOption, const char *sValue, char *cfg_var, unsigned nMaxSize)
{
	int nResult = -1;

	if (strlen(sValue) >= nMaxSize)
		log_msg(FM_LOG_ERR, "config: Value for option %s is too long", sOption);
	else {
#ifdef DEBUG
		log_msg(FM_LOG_DEBUG, "config: %s = %s", sOption, sValue);
#endif
		strcpy(cfg_var, sValue);
		nResult = 0;
	}
	return nResult;
}

// Get IP value and store in config var.
int set_ip_config(const char *sOption, const char *sValue, struct in_addr *ip)
{
	int nResult = 0;
	struct in_addr tip = {0};

	if (*sValue) { // if value present
		if (!inet_aton(sValue, &tip)) {
			log_msg(FM_LOG_ERR, "config: Invalid IP address");
			nResult = -1;
		}
	}
	if (nResult == 0) {
		*ip = tip;
#ifdef DEBUG
		log_msg(FM_LOG_DEBUG, "config: %s = %s", sOption, sValue);
#endif
	}
	return nResult;
}

// Get MSN value and store in config var.
int set_msn_config(const char *sOption, const char *sValue, msn_data *cfg_msn)
{
	int nResult = -1;
	int n, nLenMSN, nLenName;
	const char *s;

	n = atoi(sOption + 3);
	if (n < 1 || n > MAX_MSN)
		log_msg(FM_LOG_ERR, "config: MSN index %d is out of range", n);
	else {
		s = strchr(sValue, ';');
		if (NULL == s)
			log_msg(FM_LOG_ERR, "config: Missing ';' separator for MSN%d", n);
		else {
			nLenMSN = s - sValue;
			nLenName = strlen(++s);
			if (nLenMSN >= sizeof(cfg_msn->sMSN))
				log_msg(FM_LOG_ERR, "config: MSN%d is too long", n);
			if (nLenName >= sizeof(cfg_msn->sName))
				log_msg(FM_LOG_ERR, "config: Name for MSN%d is too long", n);
			if (nLenMSN < sizeof(cfg_msn->sMSN) && nLenName < sizeof(cfg_msn->sName)) {
				--n;
				strncpy(cfg_msn[n].sMSN, sValue, nLenMSN);
				cfg_msn[n].sMSN[nLenMSN] = '\0';
				strcpy(cfg_msn[n].sName, s);
#ifdef DEBUG
				log_msg(FM_LOG_DEBUG, "config: %s = %s;%s", sOption, cfg_msn[n].sMSN, cfg_msn[n].sName);
#endif
				nResult = 0;
			}
		}
	}
	return nResult;
}

// Read configuration file.
int read_config(const char *sCfgFile, config_data *data)
{
	int nLine = 0;
	char sLine[256];
	const char *sOption;
	const char *sValue;
	FILE *f;

	f = fopen(sCfgFile, "r");
	if (NULL != f) {
#ifdef DEBUG
		log_msg(FM_LOG_DEBUG, "config: reading config from file %s", sCfgFile);
#endif
		do {
			if (NULL == fgets(sLine, 255, f))
				break;
			++nLine;
			sOption = parse_config(sLine, &sValue);
			if (NULL == sOption)
				log_msg(FM_LOG_ERR, "config: Parse error in config file line %d", nLine);
			else if (!strcmp(sOption, "SHOW_ALL"))
				set_bool_config(sOption, sValue, &data->nShowAll);
			else if (!strcmp(sOption, "USE_OSD"))
				set_bool_config(sOption, sValue, &data->nUseOSD);
			else if (!strcmp(sOption, "ISO_FILES"))
				set_bool_config(sOption, sValue, &data->nISO);
			else if (!strcmp(sOption, "LOG_LEVEL"))
				set_int_config(sOption, sValue, &data->nLogLevel, 0, FM_LOG_MASK);
			else if (!strcmp(sOption, "START_DELAY"))
				set_int_config(sOption, sValue, &data->nStartDelay, 0, 300);
			else if (!strcmp(sOption, "RETRY_DELAY"))
				set_int_config(sOption, sValue, &data->nRetryDelay, 1, 300);
			else if (!strcmp(sOption, "RETRY_CONNECT"))
				set_int_config(sOption, sValue, &data->nRetry, -1, 300);
			else if (!strcmp(sOption, "PHONEBOOK_FILE"))
				set_str_config(sOption, sValue, data->sPhoneBook, sizeof(data->sPhoneBook));
			else if (!strcmp(sOption, "LOG_FILE"))
				set_str_config(sOption, sValue, data->sLogFile, sizeof(data->sLogFile));
			else if (!strcmp(sOption, "IP"))
				set_ip_config(sOption, sValue, &data->ip);
			else if (!strncmp(sOption, "MSN", 3) && atoi(sOption+3) >= 1 && atoi(sOption+3) <= MAX_MSN)
				set_msn_config(sOption, sValue, data->msn);
			else if (*sOption != '\0')
				log_msg(FM_LOG_ERR, "config: Unknown option %s", sOption);
		} while (!feof(f));
		fclose(f);
		// If no MSNs are defined show all incoming calls
		if (data->msn[0].sMSN == '\0')
			data->nShowAll = 1;
	}
	return f == NULL ? -1 : 0;
}
