/* fritzmon - Show incoming telephone calls on UFS-910 with FritzBox

   Jochen Arndt <joe127@t-online.de>
   
   Based on callmonitor by mgck
   
   Requires:
   - UFS-910 with ethernet connection to FritzBox
   - Call signalisation enabled on FritzBox (dial #96*5*)
   
   Usage: fritzmon [-h] [-k] [-d <seconds>] [-f <config-file>]
    -h = Show command line help (all other options are ignored).
    -k = Kill running daemon and exit.
    -d = Sleep for <seconds> seconds before trying to connect to FritzBox.
         Use this option if fritzmon is called during system startup
         when eth0 is not up so far.
    -f = Load specified config file.
    -l = Set log (verbose) level.
    -V = Print program version and exit.
         
   Hilights:
   - Determines FritzBox IP.
   - Tries to re-connect when connection is lost (e.g. ethernet cable unplugged,
      FritzBox restarted by power off/on or after software update).
   - No need of external programs if only the front display show messages
   - Runs as daemon that can be started by /config/start.sh
    
   TODO:
   - Test re-connect operation
   - Should lock when closing socket (avoid race condition if SIGTERM signaled)
   - Add localized message strings to config file
*/

#define VERSION		"0.1.2"

// If defined, fritzmon may be called with -t "<string>" option to display string.
#define TEST_CMD

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <time.h>
#include <net/if.h>	/* interface flags, must be included before ifaddrs.h! */
#include <ifaddrs.h>

#ifdef TEST_CMD
#include <ctype.h>
#endif

#include "fritzmon.h"
#include "config.h"
#include "osd.h"
#include "vfd.h"

// Config section
// Localized message strings
// NOTE: When using non ASCII characters here, they are UTF-8 encoded!
#define MSG_UNKNOWN_CALLER	"Unbekannt"
#define MSG_CALLER		"Anruf von:\n%s (%s)"
#define MSG_NO_CALLER		"Anruf von:\n%s"
#define MSG_LINE		"\nan %s"
//#define MSG_UNKNOWN_CALLER	"Unknown"
//#define MSG_CALLER		"Call from:\n%s (%s)"
//#define MSG_NO_CALLER		"Call from:\n%s"
//#define MSG_LINE		"\nto %s"
//MSG_UNKNOWN_CALLER: Text displayed when no number present (Front display and OSD).
//MSG_CALLER:         OSD: Top of message when entry found in phone book. Callers name and number.
//MSG_NO_CALLER:      OSD: Top of message when no phone book entry. Callers number or MSG_UNKNWON.
//MSG_LINE:           OSD: Last line. Line name (if MSN configred) or line number.

// Default setting for log file.
// If not defined, it must be specified in the config file to enable logging.
#ifdef DEBUG
#define DEF_LOG_FILE	"/var/log/fritzmon.log"
#endif

// Buffer size for OSD messages
#define MSG_BUF_SIZE	128
// End of config section 

// Path and name of lock file
#define PID_FILE	"/var/run/fritzmon.pid"
// Port on FritzBox
#define FB_PORT		1012

// If defined, phonebook may contain wild card numbers like '123*'
//#define PHONEBOOK_MATCH

// socket must be global for closing in signal handler
static int fd_sock = -1;

// configuration data
static config_data cfg_data = {
	{0},			// .ip           use auto-config for IP
	2,			// .nLogLevel    log level = warning
	1,			// .nShowAll     show all incoming calls reagrdless of MSN
	1,			// .nShowMSN     always display MSN (numeric if not configured)
	0,			// .nUseOSD      don't use OSD shell for output
	VFD_CONVERT_UTF8,	// .nISO         files are UTF-8 encoded, not ISO 8859-1
	0,			// .nStartDelay  no start-up delay
    	10,			// .nRetryDelay  re-connect interval is 10 seconds
	30,			// .nRetry       max. 30 re-connect tries
	"",			// .sPhoneBook   no phone book
#ifdef DEF_LOG_FILE
	DEF_LOG_FILE		// .sLogFile     
#else
	""			// .sLogFile     no log file
#endif
};

void log_msg(int nLevel, const char *sMsg, ...)
{
	int nErr = 0;
	int nCon = 0;
	char sBuf[128];
	char sMsgBuf[128];
	const char *sPrefix = "";
	va_list p;
	FILE *f;
		    
	if ((nLevel & FM_LOG_MASK) <= cfg_data.nLogLevel) {
		if (nLevel & FM_LOG_SYSERR)
			nErr = errno;
		if (nLevel & FM_LOG_CON)
		        nCon = 1;
		nLevel &= FM_LOG_MASK;
		switch (nLevel) {
		case FM_LOG_ERR : sPrefix = " error"; break;
		case FM_LOG_WARN : sPrefix = " warning"; break;
		case FM_LOG_INFO : sPrefix = " info"; break;
		default : sPrefix = " debug";
		}
		va_start(p, sMsg);
		vsnprintf(sBuf, 128, sMsg, p);
		va_end(p);
		sBuf[127] = '\0';
		if (nErr) {
			// append system error
			snprintf(sMsgBuf, 128, "fritzmon%s: %s: %s (errno %d)", sPrefix, sBuf, strerror(nErr), nErr);
		}
		else
			snprintf(sMsgBuf, 128, "fritzmon%s: %s", sPrefix, sBuf);
		sMsgBuf[127] = '\0';
		if (nCon)
			fprintf(nLevel <= FM_LOG_WARN ? stderr : stdout, "%s\n", sMsgBuf);
		if (cfg_data.sLogFile[0]) {
			f = fopen(cfg_data.sLogFile, "ab");
			if (NULL != f) {
				time_t t = time(NULL);
				struct tm *tm;
				tm = gmtime(&t);
				fprintf(f, "%4d-%02d-%02d %02d:%02d:%02d %s\n",
				    1900 + tm->tm_year, 1 + tm->tm_mon, tm->tm_mday,
				    tm->tm_hour, tm->tm_min, tm->tm_sec, sMsgBuf);
				fclose(f);
			}
		}
	}
}

/* Create and display message */
void printmsg(const char *sNumber, const char *sMSN)
{
	int i = 0;
	int iMatch;
	const char *sCaller = NULL;
	const char *sLine = NULL;
	const char *sMatch;
	char *sEnd;
#ifdef PHONEBOOK_MATCH
	const char *sWild;
#endif
	char sBuf[128];
	char sMsg[MSG_BUF_SIZE];
	FILE *fp;

	// Get line description for destination MSN
	while (cfg_data.msn[i].sMSN[0] && i < MAX_MSN) {
		if (!strcmp(sMSN, cfg_data.msn[i].sMSN)) {
			sLine = cfg_data.msn[i].sName;
			break;
		}
		++i;
	}
	// If all calls should be shown or call to configured MSN
	if (cfg_data.nShowAll || sLine != NULL) {
		// display MSN if no name found
		if (cfg_data.nShowMSN && NULL == sLine)
			sLine = sMSN;
    		if (!strcmp(sNumber, "POTS")) {
			sNumber = MSG_UNKNOWN_CALLER;
		}
    		else if (cfg_data.sPhoneBook[0]) {
			if (NULL != (fp = fopen(cfg_data.sPhoneBook, "r"))) {
				// Look if we have a phone book entry for the number
				do {
					if (NULL == fgets(sBuf, MSG_BUF_SIZE, fp))
						break;
					remove_comment(sBuf, &sEnd);
					iMatch = 0;
					sMatch = strtok(sBuf, ";");
					if (!strcmp(sNumber, sMatch))
						iMatch = 1;
#ifdef PHONEBOOK_MATCH
					else if (NULL != (sWild = strchr(sMatch, '*')) &&
						!strncmp(sNumber, sMatch, sWild - sMatch))
						iMatch = 1;
#endif
					if (iMatch) {
						sCaller = strtok(NULL, ";");
						if (sCaller && (*sCaller == '\'' || *sCaller == '"'))
							++sCaller;
						if (NULL != sEnd)
							*sEnd = '\0';
						break;
					}
				} while (!feof(fp));
				fclose(fp);
			}
		}
		// Pass text to OSD shell
		if (cfg_data.nUseOSD) {
			i = 0;
			if (sCaller != NULL)
		    		i = snprintf(sMsg, MSG_BUF_SIZE - 1, MSG_CALLER, sCaller, sNumber);
			else
				i = snprintf(sMsg, MSG_BUF_SIZE - 1, MSG_NO_CALLER, sNumber);
			if (i < MSG_BUF_SIZE - 1 && sLine)
				i += snprintf(sMsg + i, MSG_BUF_SIZE - i, MSG_LINE, sLine);
			if (i > MSG_BUF_SIZE - 1)
				i = MSG_BUF_SIZE - 1;
			sMsg[i] = '\0';
			popup_osd(cfg_data.nISO, sMsg);
		}
		// Show text on front display (actually limited to 64 chars)
		if (sCaller != NULL)
    			snprintf(sMsg, 64, "%s %s", sCaller, sNumber);
		else
			snprintf(sMsg, 64, sNumber);
		sMsg[64] = 0;
		log_msg(FM_LOG_INFO, sMsg);
		vfd_write(cfg_data.nISO, -1, sMsg);
	}
}

/* Write PID to lock file. */
int create_pid_file(void)
{
	FILE * f_pid;
	
	if (NULL != (f_pid = fopen(PID_FILE, "wb"))) {
		fprintf(f_pid, "%d\n", getpid());
    		fclose(f_pid);
        }
        else
		log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "open PID file");
	return f_pid == NULL ? -1 : 0;
}
						
/* Read PID from lock file. */
pid_t read_pid_file(void)
{
	pid_t pid = 0;
	char sBuf[16];
	FILE *f_pid;
        	
	if (NULL != (f_pid = fopen(PID_FILE, "rb"))) {
    		if (fgets(sBuf, 16, f_pid))
        		pid = (pid_t)atoi(sBuf);
    		fclose(f_pid);
	}
	return pid;
}

/* Get IP for secified interface name. 
   Returns IP 0.0.0.0 upon errors.
*/
struct in_addr get_ip(const char *sIfName)
{
	int is_up = 0;
	struct in_addr tIP = {0};
	struct ifaddrs *ifap = NULL;
	struct ifaddrs *ifp;
	    
	if (getifaddrs(&ifap))
		log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "getifaddrs");
	else {
    		ifp = ifap; // must use a copy of the pointer
		while (ifp) {
			if (ifp->ifa_addr && 
			    ifp->ifa_addr->sa_family == AF_INET && 
			    !strcmp(ifp->ifa_name, sIfName)) {
				is_up = 1;
				if ((ifp->ifa_flags & (IFF_UP | IFF_RUNNING)) == (IFF_UP | IFF_RUNNING)) {
					is_up = 2;
					if (ifp->ifa_addr->sa_data)
						// ifa_addr is of type sockaddr_in for famliy AF_INET
						tIP = ((struct sockaddr_in *)ifp->ifa_addr)->sin_addr;
				}
				break; // found interface
			}
			ifp = ifp->ifa_next; // next interface
		}
		freeifaddrs(ifap);
    		if (is_up == 0) // interface not found
			log_msg(FM_LOG_ERR, "Interface %s not found", sIfName);
    		else if (is_up == 1) // interface found but is not up or not running
			log_msg(FM_LOG_ERR, "Interface %s is not up or not running", sIfName);
		else if (!tIP.s_addr) // interface up and running but no IP address
			log_msg(FM_LOG_ERR, "Can't determine IP address for %s", sIfName);
#ifdef DEBUG
		else
			log_msg(FM_LOG_DEBUG, "IP address is %s", inet_ntoa(tIP));
#endif
	}
	return tIP;
}

void close_socket(void)
{
	// TODO: should lock these commands!
	if (fd_sock != -1) {
		close(fd_sock);
		fd_sock = -1;
	}
}

/* Connect to FritzBox 
   Retries connect every nRetryDelay seconds for nRetry times if nRetry > 0
   Retries infinite every nRetryDelay seconds if nRetry < 0
*/
int connect_fb(void)
{
	int nConnect = -1;
	int nRetry;
	struct sockaddr_in addr_fb;

//	close_socket();
	fd_sock = socket(AF_INET, SOCK_STREAM, 0);
	if (fd_sock == -1)
    	    log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "socket");
	else {
		nRetry = cfg_data.nRetry;
		if (nRetry > 0)
			++nRetry;
		do {
			if (nRetry > 0)
				--nRetry;
    			addr_fb.sin_family = AF_INET;
//			addr_fb.sin_addr.s_addr = cfg_data.ip.s_addr;
			addr_fb.sin_addr = cfg_data.ip;
//			addr_fb.sin_port = htons(cfg_data.nPort);
			addr_fb.sin_port = htons(FB_PORT);
			nConnect = connect(fd_sock, (struct sockaddr *)&addr_fb, sizeof(struct sockaddr_in));
			if (nConnect && nRetry)
				sleep(cfg_data.nRetryDelay);
		} while (nConnect && nRetry);
		if (!nConnect) 
			log_msg(FM_LOG_INFO, "Connected to FritzBox at %s", inet_ntoa(cfg_data.ip));
		else {
			log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "connect");
    			close_socket();
		}
	}
	return fd_sock;
}

/* Wait for incoming calls and process them
With RING data received are:
[0] = Date and time
[1] = Event ("RING")
[2] = ID
[3] = Callers number
[4] = MSN
[5] = Service (e.g. "ISDN")
TODO:
 This process is blocked while displaying a message.
 Further data from the FritzBox are cached by the socket and received later.
 The local receive buffer size used here must be big enough to
 store all data arriving during message display to avoid out-of-sync conditions.
*/
void process_calls()
{
	int i;
	int nRead;
	char buffer[512];
	const char *msg_array[16];
	char *result;
	
	while (fd_sock != -1) {
		nRead = recv(fd_sock, buffer, sizeof(buffer), 0);
		if (nRead > 0) {
#ifdef DEBUG
			if (nRead >= sizeof(buffer))
				nRead = sizeof(buffer) - 1;
			buffer[nRead] = '\0';
			log_msg(FM_LOG_DEBUG, buffer);
#endif
			result = strtok(buffer, ";");
			i = 0;
			while (NULL != result && i < 16) {
				msg_array[i] = result;
				result = strtok(NULL, ";");
				++i;
			}
			// Incoming call
			if (!strcmp(msg_array[1], "RING"))
				printmsg(msg_array[3], msg_array[4]);
		}
		else {
			if (nRead < 0) // recv error/
				log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "recv");
			else // FritzBox has closed connection
				log_msg(FM_LOG_ERR, "FritzBox has closed connection");
			close_socket();
			// Try to connect again if retry not disabled
			if (cfg_data.nRetry)
				connect_fb();
		}
	}
}
							
/* Signal function
   Use only save functions inside this handler (see man 2 signal).
*/
static void signal_handler(int signum)
{
	if (signum == SIGTERM) {
//		puts("fritzmon: Terminating.");
		if (fd_sock != -1)
		    close(fd_sock);
		unlink(PID_FILE);
		_exit(0);
	}
}
    
int main (int argc, char *argv[])
{
	int i, j;
	int nExit = -1;
	int nKill = 0;
#ifdef RETRY_CMD_OPTION
	int nRetry = -2;
#endif
	int nSleep = -1;
	int nLogLevel = -1;
	char *sCfgFile = NULL;
#ifdef TEST_CMD
	const char *sTest = NULL;
#endif
	pid_t pid;
	struct in_addr tIP;

	// This is a rather simple parser with some restrictions
	// Combining options is not supported (e.g. '-kh')
	for (i = 1; i < argc; ++i) {
		if (argv[i][0] != '-') {
			log_msg(FM_LOG_ERR | FM_LOG_CON, "Invalid command line option %s", argv[i]);
			nExit = 1;
		}
		else switch (argv[i][1]) {
		case 'k' :
			nKill = 1;
			break;
		case 'd' :
			j = 2;
			if (argv[i][2] == '=')
				++j;
			if (argv[i][j])
				nSleep = atoi(argv[i] + j);
			else if (argc > ++i)
				nSleep = atoi(argv[i]);
			else {
				log_msg(FM_LOG_ERR | FM_LOG_CON, "Missing parameter for '-d' option");
				nExit = 1;
			}
			break;
#ifdef RETRY_CMD_OPTION
		case 'r' :
			j = 2;
			if (argv[i][2] == '=')
				++j;
			if (argv[i][j])
				nRetry = atoi(argv[i] + j);
			else if (argc > ++i)
				nRetry = atoi(argv[i]);
			else {
				log_msg(FM_LOG_ERR | FM_LOG_CON, "Missing parameter for '-r' option");
				nExit = 1;
			}
			break;
#endif
		case 'f' :
			j = 2;
			if (argv[i][2] == '=')
				++j;
			if (argv[i][j])
				sCfgFile = argv[i] + j;
			else if (argc > ++i)
				sCfgFile = argv[i];
			else {
				log_msg(FM_LOG_ERR | FM_LOG_CON, "Missing parameter for '-f' option");
				nExit = 1;
			}
			break;
		case 'l' :
			j = 2;
			if (argv[i][2] == '=')
				++j;
			if (argv[i][j])
				nLogLevel = cfg_data.nLogLevel = atoi(argv[i] + j);
			else if (argc > ++i)
				nLogLevel = cfg_data.nLogLevel = atoi(argv[i]);
			else {
				log_msg(FM_LOG_ERR | FM_LOG_CON, "Missing parameter for '-l' option");
				nExit = 1;
			}
			break;
#ifdef TEST_CMD
		case 't' :
			j = 2;
			if (argv[i][2] == '=')
				++j;
			if (argv[i][j])
				sTest = argv[i] + j;
			else if (argc > ++i)
				sTest = argv[i];
			else {
				log_msg(FM_LOG_ERR | FM_LOG_CON, "Missing parameter for '-t' option");
				nExit = 1;
			}
			break;
#endif
		case 'h' :
			nExit = 0;			// show usage
			break;
		case 'V' :
			printf("fritzmon version %s\n", VERSION);
			return 0;
			break;
		default :
			log_msg(FM_LOG_ERR | FM_LOG_CON, "Unknown command line option %s", argv[i]);
			nExit = 1;
		}
	}
	if (nExit != -1) {
		fprintf(nExit ? stderr : stdout, "Usage: %s [-h] [-k] [-d <seconds>] [-f <config-file>]\n", argv[0]);
//		if (nExit != 0)
			return nExit;
	}
	
	// Read configuration from file specified on command line
	if (NULL != sCfgFile) {
		if (read_config(sCfgFile, &cfg_data)) {
			log_msg(FM_LOG_ERR | FM_LOG_SYSERR | FM_LOG_CON, "open file %s", sCfgFile);
			return 1;
		}
	}
	// Read configuration from file 'fritzmon.conf' in binary dir if it exists
	else {
		sCfgFile = malloc(strlen(argv[0]) + 6);
		if (sCfgFile) {
			strcpy(sCfgFile, argv[0]);
			strcat(sCfgFile, ".conf");
			read_config(sCfgFile, &cfg_data);
			free(sCfgFile);
		}
	}
	
	// Configuration from command line (overwrites configuration file)
	if (nSleep != -1)
		cfg_data.nStartDelay = nSleep;
	if (nLogLevel != -1)
		cfg_data.nLogLevel = nLogLevel;
#ifdef RETRY_CMD_OPTION
	if (nRetry != -2)
		cfg_data.nRetry = nRetry;
#endif
#ifdef TEST_CMD
	// Test command 'fritzmon -t <string>'
	// Write string to front display and OSD and exit.
	// When using a single hex number with <string>, print 16 chars starting at num *16
	if (NULL != sTest) {
		int nConvert = VFD_CONVERT_UTF8;
		int nLen = strlen(sTest);
		char testbuf[17];
		i = toupper(*sTest);
		if (nLen == 1 && (isdigit(i) || (i >= 'A' && i <= 'F'))) {
			i = isdigit(i) ? i - '0' : i - 'A' + 0x0A;
			for (j = 0; j < 16; j++)
				testbuf[j] = (i << 4) + j;
			testbuf[16] = 0;
			nLen = 16;
			sTest = testbuf;
			nConvert = VFD_CONVERT_NONE;
		}
		vfd_write(nConvert, nLen, sTest);
		if (cfg_data.nUseOSD && nConvert != VFD_CONVERT_NONE)
			popup_osd(0, sTest);
		return 0;
	}
#endif
	// Read PID file to check if already running as daemon
	pid = read_pid_file();

	// Check for stale lock file
	if (pid && kill(pid, 0)) {
		if (errno == ESRCH) {
			log_msg(FM_LOG_WARN | FM_LOG_CON, "Found stale lock file with PID %d", pid);
			unlink(PID_FILE);
	    		pid = 0;
		}
	}
	
	// Kill daemon and exit if kill option specified
	if (nKill) {
		nExit = 0; // exit after processing this block
		if (pid) {
			if (kill(pid, SIGTERM)) {
				log_msg(FM_LOG_ERR | FM_LOG_SYSERR | FM_LOG_CON, "kill daemon");
				nExit = 1;
			}
			else {
				log_msg(FM_LOG_INFO | FM_LOG_CON, "Daemon with PID %d killed", pid);
//				pid = 0;
			}
		}
		else
			log_msg(FM_LOG_ERR | FM_LOG_CON, "No daemon running");
	}
	
	if (nExit != -1)
		return nExit;
	
	// Check if daemon already running
	if (pid) {
		log_msg(FM_LOG_ERR | FM_LOG_CON, "Already running as daemon with PID %d",  pid);
		return 1;
	}

	// Check if phone book file defined in config file exists
	if (cfg_data.sPhoneBook[0]) {
		if (access(cfg_data.sPhoneBook, R_OK)) {
			log_msg(FM_LOG_ERR | FM_LOG_SYSERR | FM_LOG_CON, "access to phone book file %s", cfg_data.sPhoneBook);
			cfg_data.sPhoneBook[0] = '\0';
		}
	}
	// use phone book file in binary dir if it exists
	else {
		if (argv[0][0] != '/') {
			if (NULL != getcwd(cfg_data.sPhoneBook, sizeof(cfg_data.sPhoneBook) - strlen(argv[0]) - 15)) {
				strcat(cfg_data.sPhoneBook, "/");
				strcat(cfg_data.sPhoneBook, argv[0]);
			}
		}
		else if (strlen(argv[0]) + 14 < sizeof(cfg_data.sPhoneBook)) 
			strcpy(cfg_data.sPhoneBook, argv[0]);
		if (cfg_data.sPhoneBook[0] != '\0') {
			char *s = strrchr(cfg_data.sPhoneBook, '/');
			if (NULL == s)
				cfg_data.sPhoneBook[0] = '\0';
			else {
				strcpy(++s, "phonebook.txt");
				if (access(cfg_data.sPhoneBook, R_OK))
					cfg_data.sPhoneBook[0] = '\0';
			}
		}
	}
#ifdef DEBUG
	if (cfg_data.sPhoneBook[0])
		log_msg(FM_LOG_DEBUG | FM_LOG_CON, "using phone book file %s", cfg_data.sPhoneBook);
#endif

#ifdef GET_OSD_PID
	if (cfg_data.nStartDelay == 0 && cfg_data.nUseOSD && !get_process_id(OSD_PROC))
		log_msg(FM_LOG_WARN | FM_LOG_CON, "OSDshell not running");
#else
	if (cfg_data.nUseOSD && access(OSD_BIN, R_OK | X_OK)) {
		log_msg(FM_LOG_ERR | FM_LOG_SYSERR | FM_LOG_CON, "access popup script %s", OSD_BIN);
		cfg_data.nUseOSD = 0;
	}
#endif
	
        // Check for write access to pid file directory.
        // Not really necesary with UFS where only root is logged in.
//        if (create_pid_file())
//		return 1;
	    
	// Daemonize
#ifdef DEBUG
	log_msg(FM_LOG_DEBUG | FM_LOG_CON, "Forking now to daemon mode");
#endif
	if (daemon(0, 1)) {
		log_msg(FM_LOG_ERR | FM_LOG_SYSERR | FM_LOG_CON, "daemonizing");
		unlink(PID_FILE);
		return 1;
	}
	
	// Write new PID to lock file after forking
#ifdef DEBUG
	log_msg(FM_LOG_DEBUG, "Writing new PID to lock file after forking");
#endif
	create_pid_file();

	// Install signal handler. Ignore errors (must kill manually).
	if (signal(SIGTERM, signal_handler) == SIG_ERR)
		log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "Signal handler installation");
	
	// Wait some time. Necessary when eth0 not yet up.
	if (cfg_data.nStartDelay) {
#ifdef DEBUG
		log_msg(FM_LOG_DEBUG, "Waiting for %d seconds before connect", cfg_data.nStartDelay);
#endif
		sleep(cfg_data.nStartDelay);
#ifdef GET_OSD_PID
		if (cfg_data.nUseOSD && !get_process_id(OSD_PROC)) {
			log_msg(FM_LOG_WARN, "OSDshell not running");
			cfg_data.nUseOSD = 0;
		}
#endif
	    
	}
	// Get IP of UFS-910 (checks also if eth0 is up)
	tIP = get_ip("eth0");
	if (tIP.s_addr != 0) {
    		if (cfg_data.ip.s_addr == 0) {
    			// Change last number of UFS IP address to '1' to get IP of FritzBox
			tIP.s_addr &= 0x00FFFFFF;
			tIP.s_addr |= 0x01000000;
    			cfg_data.ip = tIP;
		}
    		// Connect to FritzBox
		connect_fb();
		// Wait for incoming calls and process them
		process_calls();
	}

	log_msg(FM_LOG_INFO, "Terminating");
        close_socket();
	unlink(PID_FILE);
	return 1; // There was always an error when terminating here
}
