/* osd.c

   send string to OSDshell for display in popup window

*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <dirent.h>

#include "fritzmon.h"
#include "osd.h"

// Defines which character to display for unsupported one.
// Use '\0' to skip unsupported characters.
#define UNSUPPORTED_CHAR	'?'
//#define UNSUPPORTED_CHAR	' '
//#define UNSUPPORTED_CHAR	'\0'


/* Convert UTF-8 string.
   Space-like characters (TAB, VTAB, FF, LF, CR) are replaced by a space.
   Control characters (codes < 0x20 except the above ones) are ignored.
*/
unsigned char * convert_utf(const unsigned char *sMsg)
{
	unsigned i, j, nLen;
	unsigned char cIn, cOut;
	unsigned char *s;
    
	nLen = strlen(sMsg);
	s = (unsigned char *)malloc(nLen+1);
	if (NULL == s) 
		return s;
	for (i = 0, j = 0; i < nLen; i++) {
		cIn = sMsg[i];
		cOut = UNSUPPORTED_CHAR;
		if ((cIn & 0xF0) == 0xF0) // 4 byte UTF-8 character
			i += 3;
		else if ((cIn & 0xE0) == 0xE0) // 3 byte UTF-8 character
			i += 2;
		else if ((cIn & 0xC0) == 0xC0) { // 2 byte UTF-8 character
			++i;
			if (cIn == 0xC3) {
				switch (sMsg[i]) {
				case 0xA4 : s[j++] = 'a'; cOut = 'e'; break; // ä
	    			case 0xB6 : s[j++] = 'o'; cOut = 'e'; break; // ö
				case 0xBC : s[j++] = 'u'; cOut = 'e'; break; // ü
	    			case 0x84 : s[j++] = 'A'; cOut = 'e'; break; // Ä
				case 0x96 : s[j++] = 'O'; cOut = 'e'; break; // Ö
	    			case 0x9C : s[j++] = 'U'; cOut = 'e'; break; // Ü
				case 0x9F : s[j++] = cOut = 's'; break; // ß
				}
			}
		}
		else if (cIn >= ' ') // 1 byte UTF-8 character (ASCII)
			cOut = cIn;
		else if (isspace(cIn)) // replace space-like by space
			cOut = ' ';
		else // ignore control codes
			cOut = '\0';
		if (cOut)
			s[j++] = cOut;
	}
	s[j] = 0;
	return s;
}

// Convert ISO 8859-1 string
unsigned char * convert_iso(const unsigned char *sMsg)
{
	unsigned i, j, nLen;
	unsigned char cIn, cOut;
	unsigned char *s;
    
	nLen = strlen(sMsg);
	s = (unsigned char *)malloc((nLen << 1) + 1);
	if (NULL == s) 
		return s;
	for (i = 0, j = 0; i < nLen; i++) {
		cIn = sMsg[i];
		cOut = UNSUPPORTED_CHAR;
		if (cIn >= 0x80) {
			switch (cIn) {
			case 0xE4 : s[j++] = 'a'; cOut = 'e'; break; // ä
	    		case 0xF6 : s[j++] = 'o'; cOut = 'e'; break; // ö
			case 0xFC : s[j++] = 'u'; cOut = 'e'; break; // ü
	    		case 0xC4 : s[j++] = 'A'; cOut = 'e'; break; // Ä
			case 0xD6 : s[j++] = 'O'; cOut = 'e'; break; // Ö
	    		case 0xDC : s[j++] = 'U'; cOut = 'e'; break; // Ü
			case 0xDF : s[j++] = cOut = 's'; break; // ß
//			default : cOut = UNSUPPORTED_CHAR;
			}
		}
		else if (cIn >= ' ') // ASCII character
			cOut = cIn;
		else if (isspace(cIn)) // replace space-like by space
			cOut = ' ';
		else // ignore control codes
			cOut = '\0';
		if (cOut)
			s[j++] = cOut;
	}
	s[j] = 0;
	return s;
}

#ifdef GET_OSD_PID
// Get pid of any process.
// Iterates through /proc and checks pid subdirs.
// CAUTION: 
//  May fail if the process name is passed as argument to other processes!
int get_process_id(const char * sProcess)
{
	static int ppid = 0;
    
	int pid, nLen, fd;
	DIR *proc;
        struct dirent *dir;
	char sCmdLine[24];
        char sBuf[64];
	const char *s;
    
	// check if previously determined pid is still valid
	if (ppid && kill(ppid, 0)) {
		log_msg(FM_LOG_INFO, "Ignoring stale %s pid %d", sProcess, ppid);
		ppid = 0;
	}
        if (ppid == 0) {
#ifdef DEBUG
		log_msg(FM_LOG_DEBUG, "Searching for pid of %s", sProcess);
#endif
    	        nLen = strlen(sProcess);
		proc = opendir("/proc");
		while (ppid == 0 && NULL != (dir = readdir(proc))) {
			pid = atoi(dir->d_name);
			if (pid && dir->d_type == DT_DIR) {
				sprintf(sCmdLine, "/proc/%d/cmdline", pid);
				if ((fd = open(sCmdLine, O_RDONLY)) != -1) {
					if (read(fd, sBuf, 64) > 0) {
						s = strstr(sBuf, sProcess);
						if (s && s[nLen] <= ' ')
							ppid = pid;
					}
					close(fd);
				}
			}
		}
#ifdef DEBUG
		if (ppid)
			log_msg(FM_LOG_DEBUG, "pid of %s is %d", sProcess, ppid);
		else
			log_msg(FM_LOG_DEBUG, "no pid found for %s", sProcess);
#endif
	}
	return ppid;
}

// popup using OSD shell:
//  copy text to file 'tmp/output'
//  send SIGUSR1 to OSDshell: kill(pid, SIGUSR1)
void popup_osd(int nISO, const char *msg)
{
	int fd;
	pid_t pid;
	char *s;
	
	if ((pid = get_process_id(OSD_PROC)) != 0) {
		if (nISO)
			s = convert_iso(msg);
		else
			s = convert_utf(msg);
		if (NULL != s) {
			if ((fd = open("/tmp/output", O_CREAT | O_TRUNC | O_WRONLY)) != -1) {
				write(fd, msg, strlen(msg));
				close(fd);
				kill(pid, SIGUSR1);
			}
			free(s);
		}
	}
}

#else

// popup using popup script
void popup_osd(int nISO, const char *msg)
{
	int   nStatus;
	pid_t pid = 0;
	char *s;

	if (nISO)
		s = convert_iso(msg);
	else
		s = convert_utf(msg);
	if (NULL != s) {
		pid = fork();
		if (pid < 0) // fork failed
			log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "fork");
		else if (pid == 0) { // child
			execlp(OSD_BIN, OSD_BIN, msg , (char*)NULL);
			// only reached if execlp fails
			log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "exec");
		}
		else { // parent
			// Wait until message has been displayed.
			// If another call comes in meanwhile, it is processed afterwards.
			if (wait(&nStatus) == -1)
				log_msg(FM_LOG_ERR | FM_LOG_SYSERR, "wait");
		}
		free(s);
	}
}
#endif

